#include "config.h"
#include "bloom.h"

#include <openssl/md5.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <fcntl.h>

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <set>
#include <vector>
#include <map>

#ifdef HAVE_INTTYPES_H
#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS
#endif
#include <inttypes.h>
#endif


#include "discriminator.h"
#include "block_scan.h"
#include "aftimer.h"
#include "utils.h"


#ifndef O_BINARY
#define O_BINARY 0
#endif

#define OPT_T_DEFAULT 600
#define OPT_E_DEFAULT 0.01

const int read_size = 65536;
const int initial_reads_per_second = 200;
const int block_size = 4096;

const char *progname = 0;
using namespace std;
int opt_debug = 0;
int  opt_M = 28;
int  opt_k = 4;


/**
 * votemap_t keeps track of how many votes each particular vote type gets.
 */
class votemap_t:public map<string,int> {
};


class block_scanner {
public:
    block_scanner():fd(0),filesize(0),seen_blocks(0){}
    ~block_scanner(){
	if(fd>0) close(fd);
    }
    int open(const char *fn){
	fd = ::open(fn,O_RDONLY|O_BINARY,0);
	if(fd>0) filesize = get_filesize(fd);
	return fd;
    }
    void process_buffer(const sdbuf &cbuf);
    int process_block_number(int64_t block);
    int process_random(int seconds,double desired_error_rate);
    int process_all();
    int fd;
    int64_t filesize;
    votemap_t votes;
    votelist_t notable_votes;
    int64_t seen_blocks;
    void print_stats();
};


void block_scanner::process_buffer(const sdbuf &buf)
{
    /* Run all of the discriminators and get the highest vote */
    votelist_t block_votes;
    for(discriminator_list::iterator id = discriminators.begin(); id!=discriminators.end();id++){
	votelist_t r = (*id)->identify(buf);
	if(opt_debug>1) cout << (*id)->name() << " returned " << r.size() << " results \n" << r;
	for(votelist_t::const_iterator it = r.begin();it!=r.end();it++){
	    block_votes.push_back(*it);	/* Store the votes for this block */
	    if((*id)->notable_votes()) notable_votes.push_back(*it); // store notable votes
	}
    }
    const vote *best = block_votes.best_vote();
    if(best){
	if(opt_debug) cout << buf.block_number << " BEST: " << *best << "\n";
	votes[best->name] = votes[best->name] + 1;
    }
    seen_blocks++;
}


/* Process a block; return 0 if successful, -1 if error. */
   
int block_scanner::process_block_number(int64_t block)
{
    if(opt_debug>1) cerr << "process_block_number(" << block << ")\n";
    int64_t before_blocks = 0;
    ssize_t buflen = block_size+block_size;
    int64_t pos = (block-before_blocks)*block_size;

    sdbuf buf(buflen);
    lseek(fd,pos,SEEK_SET);
    ssize_t rsize = read(fd,buf.buf,buf.buflen);
    if(rsize==buflen){
	buf.block_number = block;
	buf.block_size   = block_size;
	buf.calc_block_histogram();
	process_buffer(buf);
	return 0;
    } 
    cerr << "read " << buflen << " bytes at offset "
	 << pos <<  " = " << rsize << " : " << strerror(errno) << "\n";
    return -1;
}

/* Return a random floating point number between 0 and 1 */
double randomFP()
{
    const double bigint = (double)18446744073709551616.0;
    return (double)((((int64_t)random())<<32)|random()) / bigint; 
}

int block_scanner::process_random(int seconds,double desired_error_rate)
{
    int64_t total_blocks_scanned = 0;
    int64_t N = filesize / block_size;
    int reads_per_second = initial_reads_per_second;
    set<int64_t>blocks_previously_read;
    long double error_rate = 1.0;
    aftimer elapsed;
    double M = 1024*1024*10 / block_size;

    if(opt_debug){
	cout << "N=" << N << "\n";
	cout << "M=" << M << "\n";
    }

    elapsed.start();
#ifdef HAVE_SRANDOMDEV
    srandomdev();
#endif
    do {
	aftimer t;

	t.start();

	/* Come up with a list of blocks to read */
	set<int64_t>blocks_to_read;
	if(reads_per_second>N){
	    fprintf(stderr,"N=%"PRId64", but reads_per_block=%d\n",
		    N,reads_per_second);
	    exit(1);
	}
	for(int n=0;n<reads_per_second;n++){
	    int counter = 0;
	    int64_t random_block=0;
	    do {
		double R = randomFP();
		random_block = (int64_t)(R * N);
		if(counter++ > reads_per_second*10){
		    /* There seem to be no more blocks to read;
		     * break out of inner and outer loops.
		     */
		    n = reads_per_second; 
		    break; 
		}
	    } while(blocks_to_read.find(random_block)!=blocks_to_read.end());
	    blocks_to_read.insert(random_block);
	}

	/* Now read the randomly chosen blocks */
	for(set<int64_t>::const_iterator it = blocks_to_read.begin();
	    it!=blocks_to_read.end();it++){
	    process_block_number(*it);
	    total_blocks_scanned++;
	    error_rate *= (double)((N-(total_blocks_scanned-1)) - M) /
		(N-(total_blocks_scanned-1));
	}

	t.stop();
	int blocks_per_second = reads_per_second / t.elapsed_seconds();
	cout << "Blocks Scanned: " << total_blocks_scanned << "  error rate: " << error_rate
	     << "  blocks/sec: " << blocks_per_second << "\n";
    } while(elapsed.elapsed_seconds()<seconds && error_rate > desired_error_rate);
    return 0;
}

int block_scanner::process_all()
{
    for(int64_t block = 0;block<filesize/block_size;block++){
	process_block_number(block);
    }
    return 0;
}

void block_scanner::print_stats()
{
    cout << "Total blocks sampled: " <<seen_blocks << " ("
	 << (double)seen_blocks *100.0 / ((double)filesize/block_size) << "%)\n";
    if(seen_blocks==0) return;
    for(votemap_t::const_iterator it = votes.begin(); it!=votes.end(); it++){
	int64_t count = (*it).second;
	double frac = (double)(*it).second / seen_blocks;
	cout << (*it).first << " count:" << count << " fraction: " << frac*100.0 << "%\n";
    }
    if(notable_votes.size()){
	cout << "Notable Blocks:\n";
	for(votelist_t::const_iterator it = notable_votes.begin(); it!=notable_votes.end(); it++){
	    cout << (*it) << "\n";
	}
    }
}

void usage()
{
    cerr << "Usage: block_scan [options] file_or_device\n";
    cerr << "Random Sampling Mode options:\n";
    cerr << " -t NN  random scan for a maximum of NN seconds (default "<<OPT_T_DEFAULT <<")\n";
    cerr << " -e PP  random scan until error rate is less than PP (default "<<OPT_E_DEFAULT <<")\n";
    cerr << "Full Disk Evaluation Mode options:\n";
    cerr << " -a     scan the entire disk, beginning to end\n";
    cerr << "Options common to Random Sampling and Full Disk Evaluation:\n";
    cerr << " -b fn  adds a bloom filter for the analysis\n";
    cerr << " -B     Only run Bloom filter\n";
    cerr << "Bloom Filter Creation Mode:\n";
    cerr << " -c fn  Create a bloom filter file using file as input\n";
    cerr << " -k NN  Specifies k (default is " << opt_k << ") \n";
    cerr << " -M NN  Specifies M (default is " << opt_M << ") \n";
    cerr << "\n";
    cerr << "Debugging Options:\n";
    cerr << " -d     debug; print everything as you do it.\n";
    cerr << " -v     print version number, built-in discriminators, and exit.\n";
    cerr << "\n";
    exit(1);
}

bool allzero(const u_char *buf,size_t bufsize)
{
    for(size_t i=0;i<bufsize;i++){
	if(buf[i]) return false;
    }
    return true;
}

void bf_create(int argc,char **argv,const char *fn,int opt_M,int opt_k)
{
    NSRLBloom bf;

    /* See if the bloom filter currently exists */
    if(access(fn,R_OK)==0){
	if(bf.open(fn,O_RDWR)==0){
	    printf("Opened existing bloom filter %s; M=%d (%dMiB), ki=%d\n",
		   fn,bf.M,(1<<(bf.M-3))/(1024*1024),bf.k);
	} else {
	    fprintf(stderr,"Cannot open existing bloom filter file %s\n",fn);
	    exit(1);
	}
    } else {
	if(bf.create(fn,128,opt_M,opt_k,"created by block_scan")==0){
	    printf("Creating Bloom Filter '%s' with M=%d (%dMiB) and k=%d \n",
		   fn,opt_M,(1<<(opt_M-3))/(1024*1024),opt_k);
	} else {
	    perror("bloom create failed\n");
	    exit(1);
	}
    }

    printf("Building sector hashes from %d files...\n",argc);
    map<string,int>hashes;
    while(*argv){
	printf("Hashing file %s...",*argv);
	if(opt_debug) putchar('\n');
	fflush(stdout);
	int fd = open(*argv,O_RDONLY | O_BINARY);
	if(fd<0){
	    perror(*argv);
	    continue;
	}
	sdbuf buf(block_size);
	int64_t total=0;
	int64_t unique=0;
	while(read(fd,buf.buf,block_size)==block_size){
	    if(buf.null_block()==false){
		u_char md5hash[16];
		buf.calc_md5(md5hash);
		string hash_string((char *)md5hash,16);
		total++;
		hashes[hash_string]++;
		if(hashes[hash_string]==1) unique++;
		if(opt_debug){
		    char md5hex[33];
		    nsrl_hexbuf(md5hex,sizeof(md5hex),md5hash,16,0);
		    printf("   %s\n",md5hex);
		}
	    }
	}
	printf("%"PRId64" hashes (%"PRId64" unique; %"PRId64" dups)\n",total,unique,total-unique);
	argv++;
	argc--;
    }
    printf("Adding unique hashes to bloom filter...\n");
    for(map<string,int>::const_iterator it = hashes.begin(); it!=hashes.end();it++){
	bf.add((const u_char *)(*it).first.data());
    }
    bf.print_info();
    bf.print_usage();
    exit(1);
    
}

int main(int argc,char **argv)
{
    const char *opt_create  = 0;
    int opt_all = 0;
    int opt_t   = OPT_T_DEFAULT;
    double opt_e   = OPT_E_DEFAULT;
    int opt_bloom_only=0;
    vector<string> bloom_list;
    progname = argv[0];
    setvbuf(stdout,0,_IONBF,0);		// don't buffer stdout
    int ch;
    discriminator_list::init();
    while ((ch = getopt(argc, argv, "adb:c:k:M:t:Be:v")) != -1) {
	switch (ch) {
	case 'B': opt_bloom_only=1;break;
	case 'a': opt_all = 1;break;
	case 'd': opt_debug++;break;
	case 'b': bloom_list.push_back(optarg); break;
	case 'c': opt_create = optarg; break;
	case 'M': opt_M = atoi(optarg); break;
	case 'k': opt_k = atoi(optarg); break;
	case 't': opt_t = atoi(optarg); break;
	case 'e': opt_e = atof(optarg); break;
	case 'v':
	    cout << "Version: " << PACKAGE_VERSION << "\n";
	    cout << "Block size: " << block_size << "\n";
	    cout << "Compiled-in discriminators:\n";
	    for(discriminator_list::const_iterator it = discriminators.begin(); it!=discriminators.end(); it++){
		cout << "  " << (*it)->name() << "\n";
	    }
	    exit(1);
	    
	default:
	case 'h': case '?':
	    usage();
	    break;
	}
    }
    argc -= optind;
    argv += optind;
    if(argc==0) usage();

    /* Are we creating a bloom file? */
    if(opt_create){
	bf_create(argc,argv,opt_create,opt_M,opt_k);
	exit(0);
    }

    if(argc!=1) usage();
    if(opt_bloom_only) discriminators.clear();

    /* Now add the bloom filters */
    for(vector<string>::const_iterator it = bloom_list.begin(); it!=bloom_list.end(); it++){
	discriminator *d = bloom_discriminator::new_bloom(*it);
	if(d) discriminators.push_back(d);
	else {
	    cerr << "Cannot open bloom file: " << *it << "\n";
	    exit(1);
	}
    }

    block_scanner ss;
    if(ss.open(argv[0])<0){
	cerr << "Cannot open " <<argv[0] << " : " << strerror(errno) << "\n";
	exit(1);
    }


    if(opt_all) ss.process_all();
    else ss.process_random(opt_t,opt_e);
    ss.print_stats();
    return(0);
}
